;==============================================================================
;PHOS - Console Component - PHOS - Console Component - PHOS - Console Component
;==============================================================================

;Phos is a bare console component you can use as a skeleton for your own
;programs in assembler or other low-level languages.
;
;Use the software by adding code to procedure phos_thread_func, which is run in a
;separate thread each time a prompt line is submitted by hitting return.

;RATIONALE
;
;When building under the windows environment, you must firstly decide whether
;to produce a Windows application (in which case you have to build your own
;console with GUI components), or a console application running in DOS-box
;The present tool is a Windows GUI application which puts up a simple to use
;text console.

;You can use/copy/share this software for any purpose you like free of charge,
;but its (co-)authors must not be held responsible for its inevitable defects.

;The source files are for MASM32 macro assembler, available at no cost from
;www.masm32.com. Read Iczelion's tutorials (web link below) for a good
;introduction to windows assembly programming.

;If you would like to get in touch with the original author,
;send e-mail to <mim@ok-schalter.de>.

;Useful links:    http://www.masm32.com (helpful/active/cool bulletin board)
;                 http://win32assembly.online.fr (Iczelion's tutorial page)

;While checking the licence status of the VGA font included, I found out that
;it's compiler, Norman De Forest, passed away.
;The web page he maintained is still up and I invite you to appreciate his
;detailed, useful contributions to the developer and user community:
;http://www.chebucto.ns.ca/~af380/Profile.html (Norman)


;==============================================================================
;PHOS  -  MASM32/WIN32 - 2009 BY MICHAEL MANGELSDORF  -  LICENSE: PUBLIC DOMAIN
;==============================================================================

.386
.model flat, stdcall   
option casemap :none  

include \masm32\include\windows.inc
include \masm32\include\user32.inc
include \masm32\include\kernel32.inc
include \masm32\include\gdi32.inc
include \masm32\include\comdlg32.inc
include \masm32\include\comctl32.inc
include \masm32\include\masm32.inc
include \masm32\include\shell32.inc
include \masm32\include\advapi32.inc

includelib \masm32\lib\user32.lib
includelib \masm32\lib\kernel32.lib
includelib \masm32\lib\gdi32.lib
includelib \masm32\lib\comdlg32.lib
includelib \masm32\lib\comctl32.lib
includelib \masm32\lib\masm32.lib
includelib \masm32\lib\shell32.lib
includelib \masm32\lib\advapi32.lib

;==============================================================================
;UNINITIALIZED DATA = UNINITIALIZED DATA = UNINITIALIZED DATA = UNINITIALIZED==
;==============================================================================

.DATA?
ALIGN 4

szText MACRO Name, Text:VARARG
        LOCAL lbl
          jmp lbl
            Name db Text,0
          lbl:
        ENDM

hWnd                            dd ?   ;application window handle
hInstance                       dd ?
CommandLine                     dd ?
ThreadID                        dd ?   ;identifies the running user thread
hphos_thread_func               dd ?   ;user thread handler
hIcon                           dd ?   ;handle for the application icon
hBackBufOldBmp                  dd ?   ;temporary store
hCharMapOldBmp                  dd ?   ;temporary store
hBackBufBmp                     dd ?   ;handle for the back buffer
hBackBufDC                      dd ?   ;device context for the back buffer
hCharMapBmp                     dd ?   ;handle for the character bitmap
hCharMapDC                      dd ?   ;device context for the character bitmap
FgCol                           dd ?   ;text foreground color
BgCol                           dd ?   ;text background color
OsCol                           dd ?   ;client region overscan color

Ps PAINTSTRUCT <?>

TempRect RECT <?>

phosThreadInfo STRUC
    hWin           DD ?
    ptrLineBuffer  DD ?
    reason         DD ?
phosThreadInfo ENDS

threadInfo phosThreadInfo <?, ?, ?>

;==============================================================================
;INITIALIZED DATA = INITIALIZED DATA = INITIALIZED DATA = INITIALIZED DATA ====
;==============================================================================

.DATA
ALIGN 4

paddingLeft                     TEXTEQU <16>
paddingTop                      TEXTEQU <16>
nLinesMax                       TEXTEQU <48>   ;number of lines in the console
nLinesStart                     TEXTEQU <17>   ;number of lines at beginning
nCols                           TEXTEQU <64>   ;number of columns in the console
colWidth                        TEXTEQU <9>    ;character width in pixels
lineHeight                      TEXTEQU <17>   ;character height in pixels

paddedClientWidth               TEXTEQU %(nCols * colWidth + 2 * paddingLeft)
paddedClientHeightMax           TEXTEQU %(nLinesMax * lineHeight + 2 * paddingTop)
paddedClientHeightMin           TEXTEQU %(nLinesStart * lineHeight + 2 * paddingTop)
bottomLine                      TEXTEQU %((nLinesMax-1) * lineHeight + paddingTop)

spaceChar                       TEXTEQU <32> ;ASCII space
WM_FINISH                       EQU WM_USER + 123

;the following three variables are used in most functions to do with
;moving the cursor. max_pos_x indicates the last character in the phrase
;the user has typed on the prompt line. cursor_pos_x holds the current position
;of the cursor on the prompt line. insert_pos_x is the position, where the next
;character typed will be inserted on the prompt line.

max_pos_x                       dd 0
cursor_pos_x                    dd 0
insert_pos_x                    dd 0

threadBusy                      dd 0       ;set to one by WM_TIMER handler
TimerID                         dd 1664    ;an arbitrary value
timer_ms                        dd 1000    ;timer interval in milliseconds

paddedClientHeightCurrent       dd paddedClientHeightMin
char_width                      dd colWidth

SizingRect RECT <0, 0, paddedClientWidth, paddedClientHeightMin>

szDemoText                      db "Thread says hello!", 0
szPhosGreet                     db "phOS Win32 CP1252/VGA", 0
szPhosHelp                      db "type 'exit' to quit", 0
szWindowCaptionPrefix           db "phOS", 0
szWindowCaption                 db 128 dup (0)

ptrCaptionPrefix                dd OFFSET szWindowCaptionPrefix

;The following two variables reserve one byte more than the number of columns.
;This is because when the point of character insertion is at the last character
;of the line, the cursor (which is displayed in front of it) needs parking.
;lineBuffer holds the character codes of the prompt line string.

szEmptyLine                     db nCols + 1 dup (spaceChar), 0
lineBuffer                      db nCols + 1 dup (spaceChar), 0

;The following define a few colors for convenience. ColOVERSCAN is the paint
;for wherever no characters appear on the client region. That isn't the same
;as BgCol (see above). Each character (FgCol) is printed on its own little
;rectangle (BgCol). The grid that appears to be separating the little
;rectangles is ColOVERSCAN.

ColBLACK                        dd 00000000h
ColBLUE                         dd 00FF0000h
ColGREEN                        dd 0000FF00h
ColCYAN                         dd 00FFFF00h
ColRED                          dd 000000FFh
ColMAGENTA                      dd 00FF00FFh
ColBROWN                        dd 00FFFF00h
ColGREY                         dd 00C0C0C0h
ColYELLOW                       dd 0000FFFFh
ColWHITE                        dd 00FFFFFFh

ColOLIVE                        dd 00002200h
ColOVERSCAN                     dd 00001400h

window_style                    dd WS_OVERLAPPEDWINDOW and not WS_MAXIMIZEBOX
window_style_ex                 dd 0 ;WS_EX_OVERLAPPEDWINDOW

;==============================================================================
;PROTOTYPES = PROTOTYPES = PROTOTYPES = PROTOTYPES = PROTOTYPES = PROTOTYPES ==
;==============================================================================

;blitcursor inverts the character at the current cursor position.
;It displays the cursor. Update cursor does not remove the cursor.
;repaint_line redisplays all characters after the current insertion point,
;implicitly removing the cursor.

blitcursor           PROTO
repaint_line         PROTO

submit_prompt        PROTO :DWORD

create_user_thread        PROTO :DWORD
thread_return        PROTO
phos_thread_func     PROTO :DWORD

blitchar             PROTO :DWORD, :DWORD, :DWORD
phos_vscroll         PROTO
blitline             PROTO :DWORD, :DWORD
phos_printstr        PROTO :DWORD, :DWORD, :DWORD

updateWindowCaption  PROTO
phos_set_colorscheme PROTO :DWORD, :DWORD, :DWORD
phos_set_caption     PROTO :DWORD

;the WinMain function is an important part of any windows application;
;important settings are made and the application window is created;
;also includes the message loop; whenever anything happens to
;the application window (mouse clicks, keypress etc), the message loop turns
;round once and calls a special function called WndProc (see below)
;to deal with the event

WinMain            PROTO :DWORD, :DWORD, :DWORD, :DWORD

;WndProc deals with OS events relevant to our application

WndProc            PROTO :DWORD, :DWORD, :DWORD, :DWORD

clear_backbuffer   PROTO ;backbuffer filled with ColOVERSCAN
charmap_setup      PROTO ;bitmap containing the character cells
backbuffer_init    PROTO ;character grid and a greeting
backbuffer_setup   PROTO ;bitmap used for double buffering

TopXY              PROTO :DWORD, :DWORD

;==============================================================================
;CODE SEGMENT = ENTRY = CODE SEGMENT = ENTRY = CODE SEGMENT = ENTRY ===========
;==============================================================================

.code
phos:

invoke GetModuleHandle, NULL
mov hInstance, eax
 
invoke GetCommandLine
mov CommandLine, eax  

invoke WinMain,hInstance, NULL, CommandLine, SW_SHOWDEFAULT

invoke ExitProcess, eax

;==============================================================================
;PROCEDURE WinMain = PROCEDURE WinMain = PROCEDURE WinMain = PROCEDURE WinMain= 
;==============================================================================

WinMain proc hInst:DWORD, hPrevInst:DWORD, CmdLine:DWORD, CmdShow:DWORD

LOCAL wc   :WNDCLASSEX
LOCAL msg  :MSG
LOCAL Wwd  :DWORD
LOCAL Wht  :DWORD
LOCAL Wtx  :DWORD
LOCAL Wty  :DWORD

    szText szClassName, "PhosWin32"
    
    mov wc.cbSize, sizeof WNDCLASSEX
    mov wc.style, CS_VREDRAW
    mov wc.lpfnWndProc, offset WndProc
    mov wc.cbClsExtra, NULL
    mov wc.cbWndExtra, NULL
    mov eax, hInst
    mov wc.hInstance, eax
    mov wc.hbrBackground, NULL
    mov wc.lpszMenuName, NULL
    mov wc.lpszClassName, offset szClassName
    
    invoke LoadIcon, hInst, 500
    mov wc.hIcon, eax
    mov wc.hIconSm, NULL

    invoke LoadCursor, NULL, IDC_ARROW
    mov wc.hCursor, eax

    invoke RegisterClassEx, ADDR wc
 
    mov eax, window_style                  ;the effective client surface
    mov ecx, window_style_ex               ;gets smaller because windows
    invoke AdjustWindowRectEx,             ;adds borders etc;
            ADDR SizingRect,               ;AdjustWindowRectEx makes up for
            eax,                           ;these losses and adjusts the
            FALSE,                         ;window size so that the final
            ecx                            ;client size matches our request

    mov ebx, OFFSET SizingRect
    mov eax, (RECT PTR [ebx]).right
    sub eax, (RECT PTR [ebx]).left
    mov ecx, (RECT PTR [ebx]).bottom
    sub ecx, (RECT PTR [ebx]).top       
    mov Wwd, eax 
    mov Wht, ecx 
 
    invoke GetSystemMetrics, SM_CXSCREEN   ;get screen width in pixels
    invoke TopXY, paddedClientWidth, eax
    mov Wtx, eax
    
    invoke GetSystemMetrics, SM_CYSCREEN   ;get screen height in pixels
    invoke TopXY, paddedClientHeightMin, eax
    mov Wty, eax

    mov eax, window_style
    mov ecx, window_style_ex
 
    invoke CreateWindowEx,
            ecx, 
            ADDR szClassName,
            ADDR szWindowCaption,
            eax, 
            Wtx,Wty,Wwd,Wht,
            NULL,NULL,
            hInst,NULL

    mov hWnd, eax

    invoke updateWindowCaption

    invoke LoadBitmap,hInstance,100
    mov hCharMapBmp, eax
        
    invoke phos_set_colorscheme,              ;default petrol green
            ColGREEN, ColOLIVE, ColOVERSCAN
    
    invoke charmap_setup
    invoke backbuffer_setup
    invoke backbuffer_init

    invoke ShowWindow, hWnd, SW_SHOWNORMAL
    invoke UpdateWindow, hWnd
  
StartLoop:

    invoke GetMessage,ADDR msg,NULL,0,0
    cmp eax, 0
    je ExitLoop
    
    invoke TranslateMessage, ADDR msg
    invoke DispatchMessage,  ADDR msg
    jmp StartLoop

ExitLoop:

    invoke SelectObject,hBackBufDC,hBackBufOldBmp
    invoke SelectObject,hCharMapDC,hCharMapOldBmp
    invoke DeleteDC,hBackBufDC
    invoke DeleteDC,hCharMapDC

    mov eax, msg.wParam
    ret

WinMain endp

;==============================================================================
;PROCEDURE WndProc = PROCEDURE WndProc = PROCEDURE WndProc = PROCEDURE WndProc= 
;==============================================================================

WndProc proc hWin:DWORD, uMsg:DWORD, wParam:DWORD, lParam:DWORD

LOCAL hDC : HDC

.if uMsg == WM_PAINT            ;this message is sent if our window was covered
                                ;by another, or if it was resized etc and needs
                                ;to be repainted
                                
    invoke InvalidateRect, hWnd, 0, 0  ;UGLY FIX - why is this necessary?                        
                                
    invoke BeginPaint,         
            hWin,               
            ADDR Ps             ;structure contains an rcPaint rectangle
                                ;with the client region to be repainted
    mov hDC, eax
                                
    mov eax, Ps.rcPaint.right
    mov ecx, Ps.rcPaint.left
    sub eax, ecx                ;EAX is width of the repaint region
    
    mov edx, Ps.rcPaint.bottom  
    mov ebx, Ps.rcPaint.top
    sub edx, ebx                ;EDX is height of the repaint region

    ;we only blit the lower nLinesCurrent of the backbuffer bitmap on screen;
    ;when the window is resized, the prompt line should stay at the bottom
                
    mov ebx, paddedClientHeightMax
    sub ebx, paddedClientHeightCurrent ;height of invisible top region
                                       ;of the backbuffer
    
    ;conceptually, the invisible top region of the backbuffer is above the
    ;client region visible on the screen; to get the backbuffer coordinate
    ;from the client region coordinate, we need to add the size of the
    ;invisible portion of the backbuffer to it 
    
    add ebx, Ps.rcPaint.top         ;add remainder to convert to backbuffer
                                    ;coordinate
                                    
    add Ps.rcPaint.top, paddingTop  ;insert the padding - shift block down
    sub edx, paddingTop             ;but must decrease size of block, too
    add ebx, paddingTop             ;and shift corresponding back buffer window
                                     
    invoke BitBlt,
            hDC,                ;target: client region of our window
            Ps.rcPaint.left,    ;to x
            Ps.rcPaint.top,     ;to y
            eax, edx,           ;width and height of pixel region
            hBackBufDC,         ;source: backbuffer
            Ps.rcPaint.left,    ;from x
            ebx,                ;from y
            SRCCOPY             ;copy

    ;the second blit copies the top padding region

    invoke BitBlt,
            hDC,                ;target: client region of our window
            0, 0,               ;target top left
            paddedClientWidth,  ;copy entire top padding
            paddingTop,
            hBackBufDC,         ;source: backbuffer
            0, 0,
            SRCCOPY             ;copy

    invoke EndPaint,
            hWin,
            ADDR Ps
                      
    mov eax, 0
    ret

.elseif uMsg == WM_SIZING                     ;user is pulling window border(s)

    ;without this message handler, the user can just resize the window
    ;however she likes - this becomes ugly quickly because during WM_PAINT,
    ;we only paint the visible portion of our back buffer bitmap;
    ;if the window is extended beyond that, we only see the desktop
    ;background
    
    ;the message sends us the newly requested coordinates of the window
    ;called the sizing rectangle;
    ;we can prevent the resizing by simply overwriting the coordinates in the
    ;sizing rectangle by what was there before
    
    ;we are essentially preventing the user from resizing horizontally, and
    ;from making the window taller than the maximum number of lines, but also
    ;showing less than the minimum number of lines
       
    ;once the sizing is over (mouse released), the WM_EXITSIZEMOVE handler
    ;is called, snapping the size of the window to the nearest line of text

    ;note that what you see here is dependent on the Windows option:
    ;"Show window contents while dragging"
    ;(right click on desktop, choose properties, appearance, effects)    

    ;*****************************
    ;handle horizontal sizing;
    ;prevent resizing by simply
    ;replacing requested size
    ;by current size
    ;*****************************
    
    mov eax, lParam
    
    mov ecx, (RECT PTR [eax]).right            
    sub ecx, (RECT PTR [eax]).left            ;ecx current sizing rect width
    
    mov edx, OFFSET SizingRect
    
    mov ebx, (RECT PTR [edx]).right
    sub ebx, (RECT PTR [edx]).left            ;ebx previous sizing rect width

    .if wParam == WMSZ_RIGHT       || \       ;pulling right border or corners?
        wParam == WMSZ_BOTTOMRIGHT || \
        wParam == WMSZ_TOPRIGHT
    
        mov ecx, (RECT PTR [eax]).left        ;current left plus
        add ecx, ebx                          ;current width = current right
        mov (RECT PTR [eax]).right, ecx       ;overwrite requested right
    
    .elseif wParam == WMSZ_LEFT       || \    ;pulling left border or corners?
            wParam == WMSZ_BOTTOMLEFT || \
            wParam == WMSZ_TOPLEFT
    
        mov ecx, (RECT PTR [eax]).right       ;current right plus
        sub ecx, ebx                          ;current width = current left
        mov (RECT PTR [eax]).left, ecx        ;overwrite requested left

    .endif
                                  
    ;*****************************
    ;handle vertical sizing
    ;prevent resizing by
    ;replacing requested size
    ;by current size
    ;*****************************
                                              
    mov ecx, (RECT PTR [eax]).bottom          ;eax lParam
    sub ecx, (RECT PTR [eax]).top             ;ecx requested window height
    
    mov ebx, (RECT PTR [edx]).bottom                             
    sub ebx, (RECT PTR [edx]).top             ;ebx current window height        
    cmp ecx, ebx                              ;check whether requested height
    jg @F                                     ;less (shrink) or
                                              ;greater (grow) than current    
                                              ;height    
    ;CASE 1   ;****************
              ;shrinking window
              ;****************
    
    mov edx, paddedClientHeightCurrent
    add edx, ecx
    sub edx, ebx                              ;result: new/projected height
    cmp edx, paddedClientHeightMin            ;would the sizing request
    jle adjust_shrinking                      ;get us into the minimum zone?
                                                                                              
    mov paddedClientHeightCurrent, edx        ;nothing to do - just update
    
    mov ecx, (RECT PTR [eax]).top
    mov ebx, (RECT PTR [eax]).bottom
    
    jmp end_sizing                                       

adjust_shrinking:                             ;we are inside the minimum zone;
                                              ;to avoid overshooting, we do
                                              ;as if the user wanted exactly
                                              ;the minimum height, not less
                                              
    sub edx, paddedClientHeightMin            ;number of pixels the request is
                                              ;inside the minimum                
                                              
    .if wParam == WMSZ_TOP      || \          ;top border downwards?
        wParam == WMSZ_TOPLEFT  || \
        wParam == WMSZ_TOPRIGHT
      
        mov ebx, (RECT PTR [eax]).top         ;adjust the top so that we are
        add ebx, edx                          ;just outside the minimum region
        mov (RECT PTR [eax]).top, ebx

    .elseif wParam == WMSZ_BOTTOM      || \   ;bottom border upwards?
            wParam == WMSZ_BOTTOMLEFT  || \
            wParam == WMSZ_BOTTOMRIGHT
      
        mov ebx, (RECT PTR [eax]).bottom      ;adjust the bottom so that we are
        sub ebx, edx                          ;just outside the minimum region
        mov (RECT PTR [eax]).bottom, ebx
             
    .endif
    
    mov edx, paddedClientHeightMin            ;finally, adjust client height
    mov paddedClientHeightCurrent, edx
    
    mov ecx, (RECT PTR [eax]).top
    mov ebx, (RECT PTR [eax]).bottom
    
    jmp end_sizing
    
    ;CASE 2   ;**************
              ;growing window
              ;**************
@@:                                              
    mov edx, paddedClientHeightCurrent
    add edx, ecx
    sub edx, ebx
    cmp edx, paddedClientHeightMax
    jge adjust_growing
    
    mov paddedClientHeightCurrent, edx        ;update
    
    mov ecx, (RECT PTR [eax]).top
    mov ebx, (RECT PTR [eax]).bottom
    
    jmp end_sizing

adjust_growing:                               ;we are outside vertical range;
                                              ;to avoid overshooting, we do
                                              ;as if the user wanted exactly
                                              ;the maximum height, not more
    
    sub edx, paddedClientHeightMax            ;number of pixels the request is
                                              ;outside the maximum

    .if wParam == WMSZ_TOP      || \          ;top edge pulled up
        wParam == WMSZ_TOPLEFT  || \
        wParam == WMSZ_TOPRIGHT
      
        mov ebx, (RECT PTR [eax]).top         ;adjust the top so that we are
        add ebx, edx                          ;just inside the vertical range
        mov (RECT PTR [eax]).top, ebx

    .elseif wParam == WMSZ_BOTTOM      || \   ;bottom edge pulled down
            wParam == WMSZ_BOTTOMLEFT  || \
            wParam == WMSZ_BOTTOMRIGHT
      
        mov ebx, (RECT PTR [eax]).bottom      ;adjust bottom so that we are
        sub ebx, edx                          ;just inside the vertical range
        mov (RECT PTR [eax]).bottom, ebx
      
    .endif

    mov edx, paddedClientHeightMax            ;finally, adjust client height
    mov paddedClientHeightCurrent, edx                                          

    mov ecx, (RECT PTR [eax]).top
    mov ebx, (RECT PTR [eax]).bottom

end_sizing:
    
    mov edx, OFFSET SizingRect
    mov (RECT PTR [edx]).top, ecx
    mov (RECT PTR [edx]).bottom, ebx
    
    invoke updateWindowCaption
    
    mov eax, 1
    ret

.elseif uMsg == WM_EXITSIZEMOVE            ;make sure that only whole lines 
                                           ;are displayed and that the window
                                           ;snaps up if resized by less
                                           ;than a whole lineHeight     
    mov ebx, OFFSET SizingRect
            
    mov eax, paddedClientHeightCurrent
    sub eax, 2 * paddingTop                  
    mov ecx, lineHeight
    mov edx, 0                    
    div ecx                                ;remainder in edx
                                           ;the remainder is the number of
                                           ;pixels needed to complete a full
                                           ;line
    mov ecx, (RECT PTR [ebx]).bottom
    sub ecx, (RECT PTR [ebx]).top          ;current height
    sub ecx, edx                           ;subtract remaining pixels
    
    sub (RECT PTR [ebx]).bottom, edx       ;update sizing rectangle
    sub paddedClientHeightCurrent, edx     ;also update client height
    
    mov edx, (RECT PTR [ebx]).right
    sub edx, (RECT PTR [ebx]).left         ;current width

    invoke SetWindowPos, hWnd,             ;snap the window if necessary
            NULL, 0, 0,
            edx, ecx,
            SWP_NOMOVE or SWP_NOZORDER
    
    mov eax, 0
    ret

.elseif uMsg == WM_FINISH       ;user thread terminated

  .if lParam == 0
   invoke thread_return
  .else
   invoke PostQuitMessage, NULL
  .endif

.elseif uMsg == WM_KEYDOWN && hphos_thread_func == 0
   
  .if wParam == VK_LEFT     ;LEFT ARROW key
    mov eax, insert_pos_x
    cmp eax, 0
    je @F
    dec cursor_pos_x
    dec insert_pos_x
    invoke blitline, OFFSET lineBuffer, 1
    
@@:
  .elseif wParam == VK_RIGHT  ;RIGHT ARROW key
    mov eax, insert_pos_x
    cmp eax, nCols - 1
    jge @F
    cmp eax, max_pos_x        ;if no chars have been typed past this point
    jge @F                    ;can't go beyond yet
    inc cursor_pos_x
    invoke blitline, OFFSET lineBuffer, 1
    inc insert_pos_x

@@:               
  .elseif wParam == VK_DELETE
    mov edi, insert_pos_x
    cmp edi, nCols
    je @F
        
    add edi, OFFSET lineBuffer
    mov esi, edi
    inc esi
    mov ecx, nCols
    sub ecx, insert_pos_x
    cld
    rep movsb
    invoke blitline, OFFSET lineBuffer, 1
    dec max_pos_x 

@@:
   .elseif wParam == VK_RETURN || wParam == VK_UP || wParam == VK_DOWN
   
     invoke submit_prompt, wParam
  
  .endif
    
.elseif uMsg == WM_CHAR && hphos_thread_func == 0                                    

  .if wParam == 8                       ;BACKSPACE key
    mov edi, insert_pos_x
    cmp edi, 0
    je @F
    
    dec insert_pos_x
    dec cursor_pos_x
    dec edi
    add edi, OFFSET lineBuffer
    mov esi, edi
    inc esi
    mov ecx, nCols
    sub ecx, insert_pos_x
    cld
    rep movsb
    invoke blitline, OFFSET lineBuffer, 1
    dec max_pos_x       
@@:    
       
  .elseif (wParam>31) && (wParam<128)   ;ALPHANUMERIC keyboard input
    mov eax, insert_pos_x               ;do not allow typing beyond
    cmp eax, nCols                      ;last column, compare to number
    jge @F                              ;of columns
    
    mov edi, nCols                      ;make room for the new character by
    add edi, OFFSET lineBuffer          ;shifting the remainder of the line
    dec edi                             ;right
    mov esi, edi
    dec esi
    mov ecx, nCols
    dec ecx
    sub ecx, eax
    std
    rep movsb    
    cld                                 ;if this is left out, the system menu
                                        ;of our window will no longer work.
    add eax, OFFSET lineBuffer
    mov edx, wParam
    mov (byte ptr [eax]), dl
    inc cursor_pos_x
    invoke blitline, OFFSET lineBuffer, 1
    inc insert_pos_x
    inc max_pos_x
 
@@:
  .endif

.elseif uMsg == WM_TIMER

    add threadBusy, 1                 ;update busy flag 
    invoke updateWindowCaption

.elseif uMsg == WM_DESTROY

    invoke PostQuitMessage, NULL
    mov eax, 0 
    ret

.endif

    invoke DefWindowProc, hWin, uMsg, wParam, lParam
    ret

WndProc endp

;==============================================================================
;PROCEDURE create_user_thread = PROCEDURE create_user_thread ==================   
;==============================================================================

;create an execution thread for the user command;start the thread and
;keep message loop running; while the thread is executing,
;keyboard input is disabled;

create_user_thread proc reason:DWORD
       
    mov ebx, OFFSET threadInfo
    
    mov eax, hWnd
    mov (phosThreadInfo PTR [ebx]).hWin, eax
    
    mov eax, OFFSET lineBuffer
    mov (phosThreadInfo PTR [ebx]).ptrLineBuffer, eax
    
    mov eax, reason
    mov (phosThreadInfo PTR [ebx]).reason, eax
   
    mov eax, OFFSET phos_thread_func 
    invoke CreateThread,
            NULL, NULL, eax,
            ebx, 0,                     ;ebx: address of threadInfo structure
            ADDR ThreadID
            
    mov hphos_thread_func, eax          ;save the thread handle
    
    invoke SetTimer, hWnd, TimerID, timer_ms, NULL         
    ret

create_user_thread endp

;==============================================================================
;PROCEDURE thread_return = PROCEDURE thread_return = PROCEDURE thread_return ==   
;==============================================================================

;user thread finished and sent a WM_FINISH message; the message handler in
;WndProc directly calls this procedure; display a fresh prompt line

thread_return proc

    mov esi, OFFSET szEmptyLine         ;clear the line buffer by
    mov edi, OFFSET lineBuffer          ;copying the empty string into it
    mov ecx, nCols
    cld
    rep movsb
    
    invoke blitline, OFFSET lineBuffer, 1

    mov eax, hphos_thread_func
    invoke CloseHandle,eax
    mov eax, 0
    mov hphos_thread_func, eax
    mov threadBusy, eax                 ;switch off busy flag
    
    invoke KillTimer, hWnd, TimerID
    invoke updateWindowCaption          ;switch off possible "busy" caption
    ret

thread_return endp

;==============================================================================
;PROCEDURE repaint_line = PROCEDURE repaint_line = PROCEDURE repaint_line =====
;==============================================================================

;invalidate the bottom line region of the client bitmap so it will be repainted
;by the WM_PAINT message handler 

repaint_line proc
    
    mov ecx, OFFSET TempRect   
    mov edx, paddedClientWidth
    mov (RECT PTR [ecx]).left, 0
    mov (RECT PTR [ecx]).top, bottomLine
    add (RECT PTR [ecx]).top, lineHeight  ;ugly fix
    mov (RECT PTR [ecx]).right, edx 
    mov (RECT PTR [ecx]).bottom, lineHeight

    invoke InvalidateRect,       ;this is to show Windows that the marked
            hWnd,                ;region has changed and needs repainting
            ecx,
            FALSE
            
    invoke UpdateWindow, hWnd    ;repaint the updated region
    ret

repaint_line endp

;==============================================================================
;PROCEDURE charmap_setup = PROCEDURE charmap_setup = PROCEDURE charmap_setup == 
;==============================================================================

;create device context and select character bitmap into it;
;the character bitmap is essentially a 2048x16 picture of the 255 characters
;we use; because we know that each cell is 8 pixels wide, we can compute
;the offset of any character cell by its index; the cell is then copied/blitted
;onto the back buffer bitmap

charmap_setup proc

LOCAL hDC : HDC

  ;following DC/bitmap used for blitting chars onto backbuffer
  
  invoke GetDC, hWnd
  mov hDC, eax
  
  invoke CreateCompatibleDC, hDC
  mov hCharMapDC, eax
  
  invoke SelectObject,hCharMapDC, hCharMapBmp
  mov hCharMapOldBmp, eax
  
  invoke ReleaseDC, hWnd, hDC

  ret
  
charmap_setup endp  

;==============================================================================
;PROCEDURE backbuffer_setup
;==============================================================================

;create device context and select back buffer bitmap into it;
;the back buffer is a copy of our client region; if we were to draw/blit
;directly onto the screen, it would flicker; using a back buffer, we first
;complete all drawing/blitting on the back buffer, and then copy the buffer
;onto the screen in one go; a very similar technique is called double-buffering

backbuffer_setup proc

LOCAL hDC : HDC

  invoke GetDC, hWnd
  mov hDC, eax
  
  invoke CreateCompatibleBitmap,    ;set up a memory bitmap as a back buffer
          hDC,                      ;for the main window, same size
          paddedClientWidth,
          paddedClientHeightMax + lineHeight      
  
  mov hBackBufBmp, eax
  
  invoke CreateCompatibleDC,        ;also need a device context
          hDC                       ;must be compatible with main window 

  mov hBackBufDC, eax
  
  invoke SelectObject,
          hBackBufDC,
          hBackBufBmp
  
  mov hBackBufOldBmp, eax

 invoke clear_backbuffer
 invoke SetBkColor, hBackBufDC, BgCol
 invoke SetTextColor, hBackBufDC, FgCol

 invoke ReleaseDC, hWnd, hDC
 
 ret
 
backbuffer_setup endp

;==============================================================================
;PROCEDURE backbuffer_init = PROCEDURE backbuffer_init ========================
;==============================================================================

;create the character grid and display a greeting message

backbuffer_init proc

    invoke blitline, ADDR szEmptyLine, 0   ;paint a line of character cells
                                             ;onto back buffer bottom line
    mov ebx, paddingTop
@@: invoke BitBlt,                           ;blit a copy of bottom line 
            hBackBufDC,                      ;up onto line n
            0, ebx,
            paddedClientWidth,
            lineHeight,
            hBackBufDC,
            0, bottomLine,
            SRCCOPY
  
    add ebx, lineHeight                      ;copy/paint next line
    cmp ebx, bottomLine
    jl @B 
    
    ;the following three calls to phos_printstr display a single line
    ;made up of three colored component strings;
    ;the final parameter is the column number; the current column number
    ;is also returned in eax so calls to phos_printstr can be chained easily
    
    invoke phos_printstr, OFFSET szPhosGreet, ColWHITE, 0
    invoke phos_vscroll
    invoke phos_printstr, OFFSET szPhosHelp, ColWHITE, 0
    invoke phos_vscroll
    
    invoke blitline, OFFSET szEmptyLine, 1
    ret
    
backbuffer_init endp    

;==============================================================================
;PROCEDURE clear_backbuffer = PROCEDURE clear_backbuffer ======================
;==============================================================================

;fill the backbuffer bitmap with the "overscan" color; this makes the client
;region appear as a uniform canvas for the grid of character cells

clear_backbuffer proc

    invoke BitBlt,
     hBackBufDC,
     0, 0,
     paddedClientWidth,
     paddedClientHeightMax,
     hBackBufDC,
     0, 0,
     BLACKNESS
    
    invoke CreateSolidBrush, OsCol 
    push eax
    
    invoke SelectObject, hBackBufDC, eax    
    push eax
    
    invoke ExtFloodFill,
            hBackBufDC,
            0,0,
            0,                     
            FLOODFILLSURFACE
    
    pop eax
    invoke SelectObject, hBackBufDC, eax
    
    pop eax
    invoke DeleteObject, eax
    
    ret

clear_backbuffer endp

;==============================================================================
;PROCEDURE blitchar = PROCEDURE blitchar = PROCEDURE blitchar ================= 
;==============================================================================

;blit a single character cell onto the back buffer bitmap in the specified
;color and at the specified character position; the lineBuffer is preserved

blitchar proc  charcode:DWORD, xpos:DWORD, color_code:DWORD

    mov ecx, xpos         ;compute the effective x-coordinate where the
                          ;character should appear on the backbuffer
    mov eax, ecx
    mul char_width            
    add eax, paddingLeft  ;add bits of padding (screen border)
    mov xpos, eax         ;update for later

    mov eax, charcode     ;compute the x offset of the character cell
                          ;within the character map
    and eax, 255          ;sanitize the range - there are only 256 characters
    shl eax, 3            ;multiply by 8, the width of each character cell
    
    push eax
    invoke SetTextColor, hBackBufDC, color_code
    pop eax
    
    invoke BitBlt,
            hBackBufDC,   ;target: backbuffer
            xpos,         ;target x
            bottomLine,   ;target y
            8, 16,        ;width, height of the character cell
            hCharMapDC,   ;source: character bitmap
            eax, 0,       ;character cell, top left
            SRCCOPY            
    
    invoke SetTextColor, hBackBufDC, FgCol
    ret

blitchar endp

;==============================================================================
;PROCEDURE phos_printstr = PROCEDURE phos_printstr = PROCEDURE phos_printstr == 
;==============================================================================

;blit a null terminated string on the bottom region of the back buffer in the
;specified color; the string is output at xpos, advancing that value by the
;length of the string and returning its final value in eax to facilitate
;subsequent calls to phos_printstr

;the string is blitted directly onto the back buffer bitmap;
;the lineBuffer is preserved

phos_printstr proc ptrlinstr:DWORD, color_code:DWORD, xpos:DWORD

    mov esi, ptrlinstr
    mov edi, 0                ;counts columns
@@:
    mov eax, 0
    mov al, [esi+edi]         ;fetch character from string, current position
    cmp al, 0                 ;check for end-of-string marker
    jne blit                  ;if not, blit the character on the backbuffer
    mov eax, xpos
    ret
     
blit:
    
    invoke blitchar, eax, xpos, color_code
    inc edi
    inc xpos
    mov eax, xpos
    cmp eax, nCols
    jl @B
    ret

phos_printstr endp

;==============================================================================
;PROCEDURE blitline = PROCEDURE blitline ====================================== 
;==============================================================================

;blit a null terminated string onto the bottom region of the back buffer;
;the lineBuffer is preserved

blitline proc ptrlinstr:DWORD, show_cursor:DWORD

    mov esi, ptrlinstr        ;null terminated string  
    mov edi, insert_pos_x     ;counts columns
@@:
    mov eax, 0
    mov al, [esi+edi]         ;fetch the character code
    cmp al, 0                 ;check for end-of-string marker
    jne blit                  ;if not, blit the character cell on the backbuffer
    mov al, 32                ;else silently convert it to space character

blit:

    invoke blitchar, eax, edi, FgCol
    inc edi
    cmp edi, nCols
    jl @B

    ;this routine also removes the cursor by overwriting it on the backbuffer;
    ;when insert_pos_x is at the last character (nCols) of lineBuffer, the
    ;cursor needs to be in front of that character position, and is actually
    ;outside the grid; the following covers this borderline case;
    
    invoke SetBkColor, hBackBufDC, ColOVERSCAN
    invoke blitchar, spaceChar, nCols, FgCol
    invoke SetBkColor, hBackBufDC, BgCol
    
    mov eax, show_cursor
    cmp eax, 0
    je @F
    invoke blitcursor
    invoke repaint_line
    
@@: ret

blitline endp

;==============================================================================
;PROCEDURE phos_vscroll = PROCEDURE phos_vscroll = PROCEDURE phos_vscroll ===== 
;==============================================================================

;copy the back buffer bitmap onto itself, effacing the topmost line; this gives
;the impression that the lines of text are scrolled up; clears the bottom line
;by blitting a blank line onto it; the lineBuffer is preserved

phos_vscroll proc
           
    invoke BitBlt,                  ;take the block of pixels consisting of all
            hBackBufDC,             ;lines but the first one, and copy it onto
            0, paddingTop,          ;the first line
            paddedClientWidth,
            (nLinesMax-1)*lineHeight,
            hBackBufDC,
            0, lineHeight + paddingTop,
            SRCCOPY
            
    invoke blitline, ADDR szEmptyLine, 0
    
    invoke InvalidateRect,          ;this is to show Windows that the marked
            hWnd,                   ;region has changed and needs repainting
            NULL,
            FALSE
            
    invoke UpdateWindow, hWnd       ;repaint the updated region
    ret

phos_vscroll endp

;==============================================================================
;PROCEDURE blitcursor = PROCEDURE blitcursor = PROCEDURE blitcursor ===========
;==============================================================================

;look up which character is at the cursor's position and blit it to back buffer
;in inverse video (foreground/background color reversed); the lineBuffer
;is preserved

blitcursor proc
    
    invoke SetBkColor, hBackBufDC, FgCol
    mov eax, cursor_pos_x
    mov ecx, eax
    add eax, OFFSET lineBuffer
    mov edx, 0    
    mov dl, (byte ptr [eax])
    
    invoke blitchar, edx, ecx, BgCol
    invoke SetBkColor, hBackBufDC, BgCol
    ret

blitcursor endp

;==============================================================================
;PROCEDURE TOPXY = PROCEDURE TOPXY = PROCEDURE TOPXY = PROCEDURE TOPXY ========
;==============================================================================

;by the MASM team - a handy procedure to calculate an initial window position

TopXY proc wDim:DWORD, sDim:DWORD
  shr sDim, 2      ; divide screen dimension by 4
  shr wDim, 2      ; divide window dimension by 2
  mov eax, wDim    ; copy window dimension into eax
  sub sDim, eax    ; sub half win dimension from half screen dimension
  mov eax, sDim
  ret
TopXY endp

;==============================================================================
;PROCEDURE submit_prompt = PROCEDURE submit_prompt = PROCEDURE submit_prompt ==
;==============================================================================

submit_prompt proc reason:DWORD

    mov eax, 0    
    mov insert_pos_x, eax               ;reset prompt line (carriage return)
    mov cursor_pos_x, eax
    mov max_pos_x, eax
    
    invoke blitline, OFFSET lineBuffer, 0
    invoke create_user_thread, reason

    ret

submit_prompt endp

;==============================================================================
;PROCEDURE updateWindowCaption = PROCEDURE UpdateWindowCaption ================   
;==============================================================================

;dynamically update the string displayed in the caption bar of the application
;window; display OK if waiting for input, and the number of seconds while the
;prompt line has been submitted and is busy

updateWindowCaption proc

    appendChar MACRO ch
        mov BYTE PTR [ebx], ch
        inc ebx
    ENDM

    szText szBusy,    " s)", 0
    szText szReady,   "(OK)", 0
        
    mov ebx, OFFSET szWindowCaption
    invoke szappend, ebx, ptrCaptionPrefix, 0   ;print phOS
    add ebx, eax
    
    appendChar spaceChar
   
    .if threadBusy == 0
        invoke szappend, ebx, OFFSET szReady, 0
    .else
        appendChar '('
        mov eax, threadBusy
        invoke dw2a, eax, ebx       
        invoke szLen, ebx
        add ebx, eax
        invoke szappend, ebx, OFFSET szBusy, 0
    .endif
 
    invoke SendMessage, hWnd, WM_SETTEXT, 0, ADDR szWindowCaption

    ret

updateWindowCaption endp

;==============================================================================
;PROCEDURE phos_set_caption = PROCEDURE phos_set_caption ========================
;==============================================================================

;set the string displayed in the caption bar of the application window

phos_set_caption proc szCaptionPrefix:DWORD

    mov eax, szCaptionPrefix
    mov ptrCaptionPrefix, eax
    ret
 
phos_set_caption endp

;==============================================================================
;PROCEDURE phos_set_colorscheme = PROCEDURE phos_set_colorscheme ==============
;==============================================================================

;set text color (foreground), text background (background) and overscan
;color of the backbuffer

phos_set_colorscheme proc fg:DWORD, bg:DWORD, os:DWORD

    mov eax, fg
    mov FgCol, eax
    
    mov eax, bg
    mov BgCol, eax
    
    mov eax, os
    mov OsCol, eax
    
    ret

phos_set_colorscheme endp

;==============================================================================
;PROCEDURE phos_thread_func = PROCEDURE phos_thread_func ======================   
;==============================================================================

;this is user specified code, executed in its own thread;
;Param is a pointer to the current lineBuffer; while the thread is executing,
;keyboard input is disabled;
;when the thread terminates, it sends a WM_FINISH message to the message queue;
;this message must be sent, so that the message handler can re-enable
;keyboard input and do some housekeeping

phos_thread_func proc infoStruc:DWORD

LOCAL prompt_line :DWORD
LOCAL hWin        :DWORD
LOCAL reason      :DWORD
    
    mov ebx, infoStruc

    mov eax, (phosThreadInfo PTR [ebx]).reason
    mov reason, eax

    mov eax, (phosThreadInfo PTR [ebx]).hWin
    mov hWin, eax
    
    ;prompt_line is zero terminated, nCols characters long,
    ;padded with spaces
    
    mov eax, (phosThreadInfo PTR [ebx]).ptrLineBuffer
    mov prompt_line, eax
    
    ;useful functions:
    ;=================
    ;phos_vscroll - scroll up by one line
    ;phos_set_colorscheme - set the colors used by the console
    ;phos_set_caption - set the window caption text
    ;phos_printstr - append strings to the prompt line   
    
    .if reason == VK_RETURN           ;user has pressed the enter key
            
      ;check if lineBuffer filled with spaces (line empty)
      invoke szCmp, OFFSET szEmptyLine, prompt_line
      
      .if eax == 0                    ;lineBuffer is not empty

        ;check if the user has entered the exit command
        szText szExitCommand, "exit", 0
        invoke InString, 1, prompt_line, OFFSET szExitCommand
        
        .if eax != 0                  ;string was found
        
          invoke Sleep, 500
          ;setting lparam of the message to 1 -> quit application
          invoke PostMessage, hWin, WM_FINISH, 0, 1
        
        .else
      
          ;invoke Sleep, 2000          ;simulate lengthy operation
                                       ;the window caption count
                                       ;the seconds elapsed
                                       ;window will still be fully
                                       ;responsive
          
          invoke phos_vscroll         ;scroll up user command    
          invoke phos_printstr,
                  OFFSET szDemoText,  ;zero terminated string
                  ColWHITE,           ;foreground color
                  0                   ;print from column zero
                  
          ;phos_printstr returns next column in eax
          inc eax
          invoke phos_printstr,
                  (phosThreadInfo PTR [ebx]).ptrLineBuffer,
                  ColYELLOW,
                  eax                 ;append to existing string
        
        .endif
        
      .endif
  
      invoke phos_vscroll
      
    .elseif eax == VK_UP            ;user has pressed the up arrow key
    .elseif eax == VK_DOWN          ;user has pressed the down arrow key
    
      ;implement some kind of BASH style command history   
    
    .endif
    
    ;must keep the following lines
    invoke PostMessage, hWin, WM_FINISH, 0, 0      ;do not delete 
    ret                                            ;do not delete

phos_thread_func endp

end phos



